import dataclasses as dc
import enum
import sys
import types
import typing as ta


##


HAS_DEFAULT_FACTORY = dc._HAS_DEFAULT_FACTORY  # type: ignore  # noqa

FIELDS_ATTR = dc._FIELDS  # type: ignore  # noqa
PARAMS_ATTR = dc._PARAMS  # type: ignore  # noqa

POST_INIT_NAME = dc._POST_INIT_NAME  # type: ignore  # noqa

Params = dc._DataclassParams  # type: ignore  # noqa


"""
@dc.dataclass(frozen=True)
class Params:
    init = True
    repr = True
    eq = True
    order = False
    unsafe_hash = False
    frozen = False
    match_args = True
    kw_only = False
    slots = False
    weakref_slot = False
"""


##


is_dataclass_instance = dc._is_dataclass_instance  # type: ignore  # noqa


##


ATOMIC_TYPES: frozenset[type]

if hasattr(dc, '_ATOMIC_TYPES'):
    ATOMIC_TYPES = getattr(dc, '_ATOMIC_TYPES')

else:
    ATOMIC_TYPES = frozenset({
        types.NoneType,
        bool,
        int,
        float,
        str,

        complex,
        bytes,

        types.EllipsisType,
        types.NotImplementedType,
        types.CodeType,
        types.BuiltinFunctionType,
        types.FunctionType,
        type,
        range,
        property,
    })


##


def _patch_missing_ctor() -> None:
    # dc.asdict uses copy.deepcopy which instantiates new _MISSING_TYPE objects which do not pass the 'foo is MISSING'
    # checks used throughout dataclasses code. Code should not depend on this behavior but it is a debugging landmine.
    if dc._MISSING_TYPE.__new__ is object.__new__:  # noqa
        def _MISSING_TYPE_new(cls):  # noqa
            return dc.MISSING
        dc._MISSING_TYPE.__new__ = _MISSING_TYPE_new  # type: ignore  # noqa


##


class FieldType(enum.Enum):
    INSTANCE = dc._FIELD  # type: ignore  # noqa
    CLASS = dc._FIELD_CLASSVAR  # type: ignore  # noqa
    INIT = dc._FIELD_INITVAR  # type: ignore  # noqa


_SELF_MODULE = None


def _self_module():
    global _SELF_MODULE
    if _SELF_MODULE is None:
        _SELF_MODULE = sys.modules[__package__.rpartition('.')[0]]
    return _SELF_MODULE


def is_classvar(cls: type, ty: ta.Any) -> bool:
    return (
        dc._is_classvar(ty, ta)  # type: ignore  # noqa
        or (isinstance(ty, str) and dc._is_type(ty, cls, ta, ta.ClassVar, dc._is_classvar))  # type: ignore  # noqa
    )


def is_initvar(cls: type, ty: ta.Any) -> bool:
    return (
        dc._is_initvar(ty, dc)  # type: ignore  # noqa
        or (
            isinstance(ty, str)
            and any(
                dc._is_type(ty, cls, mod, dc.InitVar, dc._is_initvar)  # type: ignore  # noqa
                for mod in (dc, _self_module())
            )
        )
    )


def is_kw_only(cls: type, ty: ta.Any) -> bool:
    return (
        dc._is_kw_only(ty, dc)  # type: ignore  # noqa
        or (
            isinstance(ty, str)
            and any(
                dc._is_type(ty, cls, mod, dc.KW_ONLY, dc._is_kw_only)  # type: ignore  # noqa
                for mod in (dc, _self_module())
            )
        )
    )



"""
TODO:
 - httpx catch
 - check=False
 - return non-200 HttpResponses
 - async
 - stream
"""
import abc
import http.client
import typing as ta
import urllib.error
import urllib.request

from .. import cached
from .. import dataclasses as dc
from .. import lang
from .headers import CanHttpHeaders
from .headers import HttpHeaders


if ta.TYPE_CHECKING:
    import httpx
else:
    httpx = lang.proxy_import('httpx')


##


DEFAULT_ENCODING = 'utf-8'


def is_success_status(status: int) -> bool:
    return 200 <= status < 300


##


@dc.dataclass(frozen=True)
class HttpRequest(lang.Final):
    url: str
    method: str | None = None  # noqa

    _: dc.KW_ONLY

    headers: CanHttpHeaders | None = dc.xfield(None, repr=dc.truthy_repr)
    data: bytes | str | None = dc.xfield(None, repr_fn=lambda v: '...' if v is not None else None)

    timeout_s: float | None = None

    #

    @property
    def method_or_default(self) -> str:
        if self.method is not None:
            return self.method
        if self.data is not None:
            return 'POST'
        return 'GET'

    @cached.property
    def headers_(self) -> HttpHeaders | None:
        return HttpHeaders(self.headers) if self.headers is not None else None


@dc.dataclass(frozen=True, kw_only=True)
class HttpResponse(lang.Final):
    status: int

    headers: HttpHeaders | None = dc.xfield(None, repr=dc.truthy_repr)
    data: bytes | None = dc.xfield(None, repr_fn=lambda v: '...' if v is not None else None)

    request: HttpRequest
    underlying: ta.Any = dc.field(default=None, repr=False)

    #

    @property
    def is_success(self) -> bool:
        return is_success_status(self.status)


class HttpClientError(Exception):
    @property
    def cause(self) -> BaseException | None:
        return self.__cause__


@dc.dataclass(frozen=True)
class HttpStatusError(HttpClientError):
    response: HttpResponse


class HttpClient(lang.Abstract):
    def __enter__(self) -> ta.Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def request(
            self,
            req: HttpRequest,
            *,
            check: bool = False,
    ) -> HttpResponse:
        resp = self._request(req)

        if check and not resp.is_success:
            if isinstance(resp.underlying, Exception):
                cause = resp.underlying
            else:
                cause = None
            raise HttpStatusError(resp) from cause

        return resp

    @abc.abstractmethod
    def _request(self, req: HttpRequest) -> HttpResponse:
        raise NotImplementedError


##


class UrllibHttpClient(HttpClient):
    def _request(self, req: HttpRequest) -> HttpResponse:
        d: ta.Any
        if (d := req.data) is not None:
            if isinstance(d, str):
                d = d.encode(DEFAULT_ENCODING)

        # urllib headers are dumb dicts [1], and keys *must* be strings or it will automatically add problematic default
        # headers because it doesn't see string keys in its header dict [2]. frustratingly it has no problem accepting
        # bytes values though [3].
        # [1]: https://github.com/python/cpython/blob/232b303e4ca47892f544294bf42e31dc34f0ec72/Lib/urllib/request.py#L319-L325  # noqa
        # [2]: https://github.com/python/cpython/blob/232b303e4ca47892f544294bf42e31dc34f0ec72/Lib/urllib/request.py#L1276-L1279  # noqa
        # [3]: https://github.com/python/cpython/blob/232b303e4ca47892f544294bf42e31dc34f0ec72/Lib/http/client.py#L1300-L1301  # noqa
        h: dict[str, str] = {}
        if hs := req.headers_:
            for k, v in hs.strict_dct.items():
                h[k.decode('ascii')] = v.decode('ascii')

        try:
            with urllib.request.urlopen(  # noqa
                    urllib.request.Request(  # noqa
                        req.url,
                        method=req.method_or_default,
                        headers=h,
                        data=d,
                    ),
                    timeout=req.timeout_s,
            ) as resp:
                return HttpResponse(
                    status=resp.status,
                    headers=HttpHeaders(resp.headers.items()),
                    data=resp.read(),
                    request=req,
                    underlying=resp,
                )

        except urllib.error.HTTPError as e:
            return HttpResponse(
                status=e.code,
                headers=HttpHeaders(e.headers.items()),
                data=e.read(),
                request=req,
                underlying=e,
            )

        except (urllib.error.URLError, http.client.HTTPException) as e:
            raise HttpClientError from e


##


class HttpxHttpClient(HttpClient):
    def _request(self, req: HttpRequest) -> HttpResponse:
        try:
            response = httpx.request(
                method=req.method_or_default,
                url=req.url,
                headers=req.headers_ or None,  # type: ignore
                content=req.data,
                timeout=req.timeout_s,
            )

            return HttpResponse(
                status=response.status_code,
                headers=HttpHeaders(response.headers.raw),
                data=response.content,
                request=req,
                underlying=response,
            )

        except httpx.HTTPError as e:
            raise HttpClientError from e


##


def _default_client() -> HttpClient:
    return UrllibHttpClient()


def client() -> HttpClient:
    return _default_client()


def request(
        url: str,
        method: str | None = None,
        *,
        headers: CanHttpHeaders | None = None,
        data: bytes | str | None = None,

        timeout_s: float | None = None,

        check: bool = False,

        client: HttpClient | None = None,

        **kwargs: ta.Any,
) -> HttpResponse:
    req = HttpRequest(
        url,
        method=method,

        headers=headers,
        data=data,

        timeout_s=timeout_s,

        **kwargs,
    )

    def do(cli: HttpClient) -> HttpResponse:
        return cli.request(
            req,

            check=check,
        )

    if client is not None:
        return do(client)

    else:
        with _default_client() as cli:
            return do(cli)



import dataclasses as dc
import functools
import typing as ta

from ... import lang
from ..coro import BytesSteppedCoro
from ..coro import BytesSteppedReaderCoro
from .adapters import CompressorObjectIncrementalAdapter
from .adapters import DecompressorObjectIncrementalAdapter
from .base import Compression
from .base import IncrementalCompression
from .codecs import make_compression_codec
from .codecs import make_compression_lazy_loaded_codec


if ta.TYPE_CHECKING:
    import lzma
else:
    lzma = lang.proxy_import('lzma')


##


@dc.dataclass(frozen=True, kw_only=True)
class LzmaCompression(Compression, IncrementalCompression):
    format: int | None = None

    check: int = -1
    preset: int | None = None
    filters: dict | None = None

    mem_limit: int | None = None

    def compress(self, d: bytes) -> bytes:
        return lzma.compress(
            d,
            format=self.format if self.format is not None else lzma.FORMAT_XZ,
            check=self.check,
            preset=self.preset,
            filters=self.filters,  # type: ignore[arg-type]
        )

    def decompress(self, d: bytes) -> bytes:
        return lzma.decompress(
            d,
            format=self.format if self.format is not None else lzma.FORMAT_AUTO,
            memlimit=self.mem_limit,
            filters=self.filters,  # type: ignore[arg-type]
        )

    def compress_incremental(self) -> BytesSteppedCoro[None]:
        return lang.nextgen(CompressorObjectIncrementalAdapter(
            functools.partial(  # type: ignore
                lzma.LZMACompressor,
                format=self.format if self.format is not None else lzma.FORMAT_XZ,
                check=self.check,
                preset=self.preset,
                filters=self.filters,  # type: ignore[arg-type]
            ),
        )())

    def decompress_incremental(self) -> BytesSteppedReaderCoro[None]:
        return DecompressorObjectIncrementalAdapter(
            functools.partial(  # type: ignore
                lzma.LZMADecompressor,
                format=self.format if self.format is not None else lzma.FORMAT_AUTO,
                memlimit=self.mem_limit,
                filters=self.filters,  # type: ignore[arg-type]
            ),
            trailing_error=lzma.LZMAError,
        )()


##


LZMA_CODEC = make_compression_codec('lzma', LzmaCompression)

# @omlish-manifest
_LZMA_LAZY_CODEC = make_compression_lazy_loaded_codec(__name__, 'LZMA_CODEC', LZMA_CODEC)



"""
https://github.com/pallets/werkzeug/blob/9e050f7750214d6779636813b8d661250804e811/src/werkzeug/http.py
https://github.com/pallets/werkzeug/blob/9e050f7750214d6779636813b8d661250804e811/src/werkzeug/sansio/http.py
"""
# Copyright 2007 Pallets
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1.  Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#
# 2.  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3.  Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import datetime
import re
import typing as ta
import urllib.parse

from .. import collections as col
from .. import lang
from .dates import http_date


##


_COOKIE_RE = re.compile(
    r"""
    ([^=;]*)
    (?:\s*=\s*
      (
        "(?:[^\\"]|\\.)*"
      |
        .*?
      )
    )?
    \s*;\s*
    """,
    flags=re.ASCII | re.VERBOSE,
)
_COOKIE_UNSLASH_RE = re.compile(rb'\\([0-3][0-7]{2}|.)')


def _cookie_unslash_replace(m: ta.Match[bytes]) -> bytes:
    v = m.group(1)

    if len(v) == 1:
        return v

    return int(v, 8).to_bytes(1, 'big')


def parse_cookie(
        cookie: str | None = None,
        *,
        no_latin1: bool = False,
) -> ta.MutableMapping[str, list[str]]:
    if (not no_latin1) and cookie:
        cookie = cookie.encode('latin1').decode()

    if not cookie:
        return {}

    cookie = f'{cookie};'
    out = []

    for ck, cv in _COOKIE_RE.findall(cookie):
        ck = ck.strip()
        cv = cv.strip()

        if not ck:
            continue

        if len(cv) >= 2 and cv[0] == cv[-1] == '"':
            # Work with bytes here, since a UTF-8 character could be multiple bytes.
            cv = _COOKIE_UNSLASH_RE.sub(
                _cookie_unslash_replace,
                cv[1:-1].encode(),
            ).decode(errors='replace')

        out.append((ck, cv))

    return col.multi_map(out)


##


_COOKIE_NO_QUOTE_RE = re.compile(r"[\w!#$%&'()*+\-./:<=>?@\[\]^`{|}~]*", re.ASCII)
_COOKIE_SLASH_RE = re.compile(rb'[\x00-\x19\",;\\\x7f-\xff]', re.ASCII)
_COOKIE_SLASH_MAP = {b'"': b'\\"', b'\\': b'\\\\'}
_COOKIE_SLASH_MAP.update(
    (v.to_bytes(1, 'big'), b'\\%03o' % v)
    for v in [*range(0x20), *b',;', *range(0x7F, 256)]
)


class CookieTooBigError(Exception):
    pass


def dump_cookie(
        key: str,
        value: str = '',
        *,
        max_age: datetime.timedelta | int | None = None,
        expires: str | datetime.datetime | float | None = None,
        path: str | None = '/',
        domain: str | None = None,
        secure: bool = False,
        httponly: bool = False,
        sync_expires: bool = True,
        max_size: int = 4093,
        samesite: str | None = None,
        partitioned: bool = False,
) -> str:
    if path is not None:
        # safe = https://url.spec.whatwg.org/#url-path-segment-string as well as percent for things that are already
        # quoted excluding semicolon since it's part of the header syntax
        path = urllib.parse.quote(path, safe="%!$&'()*+,/:=@")

    if domain:
        domain = domain.partition(':')[0].lstrip('.').encode('idna').decode('ascii')

    if isinstance(max_age, datetime.timedelta):
        max_age = int(max_age.total_seconds())

    if expires is not None:
        if not isinstance(expires, str):
            expires = http_date(expires)
    elif max_age is not None and sync_expires:
        expires = http_date(lang.utcnow().timestamp() + max_age)

    if samesite is not None:
        samesite = samesite.title()

        if samesite not in {'Strict', 'Lax', 'None'}:
            raise ValueError("SameSite must be 'Strict', 'Lax', or 'None'.")

    if partitioned:
        secure = True

    # Quote value if it contains characters not allowed by RFC 6265. Slash-escape with
    # three octal digits, which matches http.cookies, although the RFC suggests base64.
    if not _COOKIE_NO_QUOTE_RE.fullmatch(value):
        # Work with bytes here, since a UTF-8 character could be multiple bytes.
        value = _COOKIE_SLASH_RE.sub(
            lambda m: _COOKIE_SLASH_MAP[m.group()], value.encode(),
        ).decode('ascii')
        value = f'"{value}"'

    # Send a non-ASCII key as mojibake. Everything else should already be ASCII.
    # TODO Remove encoding dance, it seems like clients accept UTF-8 keys
    buf = [f"{key.encode().decode('latin1')}={value}"]

    for k, v in (
            ('Domain', domain),
            ('Expires', expires),
            ('Max-Age', max_age),
            ('Secure', secure),
            ('HttpOnly', httponly),
            ('Path', path),
            ('SameSite', samesite),
            ('Partitioned', partitioned),
    ):
        if v is None or v is False:
            continue

        if v is True:
            buf.append(k)
            continue

        buf.append(f'{k}={v}')

    rv = '; '.join(buf)

    # Warn if the final value of the cookie is larger than the limit. If the cookie is too large, then it may be
    # silently ignored by the browser, which can be quite hard to debug.
    cookie_size = len(rv)
    if max_size and cookie_size > max_size:
        raise CookieTooBigError(cookie_size)

    return rv



import collections
import dataclasses as dc
import types
import typing as ta

from .. import check
from .impl.metadata import METADATA_ATTR
from .impl.metadata import UserMetadata
from .impl.params import DEFAULT_FIELD_EXTRAS
from .impl.params import FieldExtras
from .impl.params import get_field_extras


T = ta.TypeVar('T')


##


def maybe_post_init(sup: ta.Any) -> bool:
    try:
        fn = sup.__post_init__
    except AttributeError:
        return False
    fn()
    return True


##


def opt_repr(o: ta.Any) -> str | None:
    return repr(o) if o is not None else None


def truthy_repr(o: ta.Any) -> str | None:
    return repr(o) if o else None


##


def fields_dict(cls_or_instance: ta.Any) -> dict[str, dc.Field]:
    return {f.name: f for f in dc.fields(cls_or_instance)}


class field_modifier:  # noqa
    def __init__(self, fn: ta.Callable[[dc.Field], dc.Field]) -> None:
        super().__init__()
        self.fn = fn

    def __ror__(self, other: T) -> T:
        return self(other)

    def __call__(self, f: T) -> T:
        return check.isinstance(self.fn(check.isinstance(f, dc.Field)), dc.Field)  # type: ignore


def chain_metadata(*mds: ta.Mapping) -> types.MappingProxyType:
    return types.MappingProxyType(collections.ChainMap(*mds))  # type: ignore  # noqa


def update_class_metadata(cls: type[T], *args: ta.Any) -> type[T]:
    check.isinstance(cls, type)
    setattr(cls, METADATA_ATTR, md := getattr(cls, METADATA_ATTR, {}))
    md.setdefault(UserMetadata, []).extend(args)
    return cls


def update_field_metadata(f: dc.Field, nmd: ta.Mapping) -> dc.Field:
    check.isinstance(f, dc.Field)
    f.metadata = chain_metadata(nmd, f.metadata)
    return f


def update_field_extras(f: dc.Field, *, unless_non_default: bool = False, **kwargs: ta.Any) -> dc.Field:
    fe = get_field_extras(f)
    return update_field_metadata(f, {
        FieldExtras: dc.replace(fe, **{
            k: v
            for k, v in kwargs.items()
            if not unless_non_default or v != getattr(DEFAULT_FIELD_EXTRAS, k)
        }),
    })


def update_fields(
        fn: ta.Callable[[str, dc.Field], dc.Field],
        fields: ta.Iterable[str] | None = None,
) -> ta.Callable[[type[T]], type[T]]:
    def inner(cls):
        if fields is None:
            for a, v in list(cls.__dict__.items()):
                if isinstance(v, dc.Field):
                    setattr(cls, a, fn(a, v))

        else:
            for a in fields:
                try:
                    v = cls.__dict__[a]
                except KeyError:
                    v = dc.field()
                else:
                    if not isinstance(v, dc.Field):
                        v = dc.field(default=v)
                setattr(cls, a, fn(a, v))

        return cls

    check.not_isinstance(fields, str)
    return inner


def update_fields_metadata(
        nmd: ta.Mapping,
        fields: ta.Iterable[str] | None = None,
) -> ta.Callable[[type[T]], type[T]]:
    def inner(a: str, f: dc.Field) -> dc.Field:
        return update_field_metadata(f, nmd)

    return update_fields(inner, fields)


##


def deep_replace(o: T, *args: str | ta.Callable[[ta.Any], ta.Mapping[str, ta.Any]]) -> T:
    if not args:
        return o
    elif len(args) == 1:
        return dc.replace(o, **args[0](o))  # type: ignore
    else:
        return dc.replace(o, **{args[0]: deep_replace(getattr(o, args[0]), *args[1:])})  # type: ignore


##


def iter_items(obj: ta.Any) -> ta.Iterator[tuple[str, ta.Any]]:
    for f in dc.fields(obj):
        yield (f.name, getattr(obj, f.name))


def iter_keys(obj: ta.Any) -> ta.Iterator[str]:
    for f in dc.fields(obj):
        yield f.name


def iter_values(obj: ta.Any) -> ta.Iterator[ta.Any]:
    for f in dc.fields(obj):
        yield getattr(obj, f.name)



"""
TODO:
 - ** can currently bind in a child/private scope shadowing an external parent binding **
 - better source tracking
 - cache/export ElementCollections lol
 - scope bindings, auto in root
 - injector-internal / blacklisted bindings (Injector itself, default scopes) without rebuilding ElementCollection
 - config - proxies, impl select, etc
  - config is probably shared with ElementCollection... but not 'bound', must be shared everywhere
  - InjectorRoot object?
 - ** eagers in any scope, on scope init/open
 - injection listeners
 - unions - raise on ambiguous - usecase: sql.AsyncEngineLike
 - multiple live request scopes on single injector - use private injectors?
"""
import contextlib
import functools
import itertools
import logging
import typing as ta
import weakref

from ... import check
from ... import lang
from ..elements import Elements
from ..exceptions import CyclicDependencyError
from ..exceptions import UnboundKeyError
from ..injector import Injector
from ..inspect import KwargsTarget
from ..keys import Key
from ..keys import as_key
from ..listeners import ProvisionListener
from ..listeners import ProvisionListenerBinding
from ..scopes import ScopeBinding
from ..scopes import Singleton
from ..scopes import ThreadScope
from ..types import Scope
from ..types import Unscoped
from .elements import ElementCollection
from .inspect import build_kwargs_target
from .scopes import ScopeImpl
from .scopes import make_scope_impl


log = logging.getLogger(__name__)


DEFAULT_SCOPES: list[Scope] = [
    Unscoped(),
    Singleton(),
    ThreadScope(),
]


class InjectorImpl(Injector, lang.Final):
    def __init__(self, ec: ElementCollection, p: ta.Optional['InjectorImpl'] = None) -> None:
        super().__init__()

        self._ec = check.isinstance(ec, ElementCollection)
        self._p: InjectorImpl | None = check.isinstance(p, (InjectorImpl, None))

        self._internal_consts: dict[Key, ta.Any] = {
            Key(Injector): self,
        }

        self._bim = ec.binding_impl_map()
        self._ekbs = ec.eager_keys_by_scope()
        self._pls: tuple[ProvisionListener, ...] = tuple(
            b.listener
            for b in itertools.chain(
                ec.elements_of_type(ProvisionListenerBinding),
                (p._pls if p is not None else ()),  # noqa
            )
        )

        self._cs: weakref.WeakSet[InjectorImpl] | None = None
        self._root: InjectorImpl = p._root if p is not None else self  # noqa

        self.__cur_req: InjectorImpl._Request | None = None

        ss = [
            *DEFAULT_SCOPES,
            *[sb.scope for sb in ec.elements_of_type(ScopeBinding)],
        ]
        self._scopes: dict[Scope, ScopeImpl] = {
            s: make_scope_impl(s) for s in ss
        }

        self._instantiate_eagers(Unscoped())
        self._instantiate_eagers(Singleton())

    _root: 'InjectorImpl'

    def _instantiate_eagers(self, sc: Scope) -> None:
        for k in self._ekbs.get(sc, ()):
            self.provide(k)

    def get_scope_impl(self, sc: Scope) -> ScopeImpl:
        return self._scopes[sc]

    def create_child(self, ec: ElementCollection) -> Injector:
        c = InjectorImpl(ec, self)
        if self._cs is None:
            self._cs = weakref.WeakSet()
        self._cs.add(c)
        return c

    class _Request:
        def __init__(self, injector: 'InjectorImpl') -> None:
            super().__init__()
            self._injector = injector
            self._provisions: dict[Key, lang.Maybe] = {}
            self._seen_keys: set[Key] = set()
            self._source_stack: list[ta.Any] = []

        def handle_key(self, key: Key) -> lang.Maybe[lang.Maybe]:
            try:
                return lang.just(self._provisions[key])
            except KeyError:
                pass
            if key in self._seen_keys:
                raise CyclicDependencyError(key)
            self._seen_keys.add(key)
            return lang.empty()

        def handle_provision(self, key: Key, mv: lang.Maybe) -> lang.Maybe:
            check.in_(key, self._seen_keys)
            check.not_in(key, self._provisions)
            self._provisions[key] = mv
            return mv

        @contextlib.contextmanager
        def push_source(self, source: ta.Any) -> ta.Iterator[None]:
            self._source_stack.append(source)
            try:
                yield
            finally:
                nsource = self._source_stack.pop()
                if source is not nsource:
                    raise Exception(f'Stack error: {source=} is not {nsource=}')

        def __enter__(self) -> ta.Self:
            return self

        def __exit__(self, *exc) -> None:
            pass

    @contextlib.contextmanager
    def _current_request(self) -> ta.Generator[_Request, None, None]:
        if (cr := self.__cur_req) is not None:
            yield cr
            return

        with self._Request(self) as cr:
            try:
                self.__cur_req = cr
                yield cr
            finally:
                self.__cur_req = None

    def _try_provide(self, key: ta.Any, *, source: ta.Any = None) -> lang.Maybe[ta.Any]:
        key = as_key(key)

        cr: InjectorImpl._Request
        with self._current_request() as cr:
            with cr.push_source(source):
                if (rv := cr.handle_key(key)).present:
                    return rv.must()

                ic = self._internal_consts.get(key)
                if ic is not None:
                    return cr.handle_provision(key, lang.just(ic))

                bi = self._bim.get(key)
                if bi is not None:
                    sc = self._scopes[bi.scope]

                    fn = lambda: sc.provide(bi, self)  # noqa
                    for pl in self._pls:
                        fn = functools.partial(pl, self, key, bi.binding, fn)
                    v = fn()

                    return cr.handle_provision(key, lang.just(v))

                if self._p is not None:
                    pv = self._p._try_provide(key, source=source)  # noqa
                    if pv.present:
                        return cr.handle_provision(key, pv)

                return cr.handle_provision(key, lang.empty())

    def _provide(self, key: ta.Any, *, source: ta.Any = None) -> ta.Any:
        v = self._try_provide(key, source=source)
        if v.present:
            return v.must()
        raise UnboundKeyError(key)

    #

    def try_provide(self, key: ta.Any) -> lang.Maybe[ta.Any]:
        return self.try_provide(key)

    def provide(self, key: ta.Any) -> ta.Any:
        v = self._try_provide(key)
        if v.present:
            return v.must()
        raise UnboundKeyError(key)

    def provide_kwargs(self, kt: KwargsTarget) -> ta.Mapping[str, ta.Any]:
        ret: dict[str, ta.Any] = {}
        for kw in kt.kwargs:
            if kw.has_default:
                if not (mv := self._try_provide(kw.key, source=kt)).present:
                    continue
                v = mv.must()
            else:
                v = self._provide(kw.key, source=kt)
            ret[kw.name] = v
        return ret

    def inject(self, obj: ta.Any) -> ta.Any:
        if isinstance(obj, KwargsTarget):
            obj, kt = obj.obj, obj
        else:
            kt = build_kwargs_target(obj)
        kws = self.provide_kwargs(kt)
        # FIXME: still 'injecting' (as in has a req) if ctor needs and uses Injector
        return obj(**kws)


def create_injector(es: Elements) -> Injector:
    return InjectorImpl(ElementCollection(es))



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
Note: string.Formatter (and string.Template) shouldn't be ignored - if they can be used they probably should be.
 - https://docs.python.org/3/library/string.html#custom-string-formatting
 - https://docs.python.org/3/library/string.html#template-strings
"""
import dataclasses as dc
import re
import typing as ta

from ..lite.check import check


@dc.dataclass(frozen=True)
class GlyphSplitMatch:
    l: str
    s: str
    r: str


class GlyphSplitter:
    def __init__(
            self,
            glyphs: ta.Tuple[str, str],
            *,
            escape_returned_doubles: bool = False,
            compact: bool = False,
    ) -> None:
        super().__init__()

        glyphs = tuple(map(check.of_isinstance(str), glyphs))  # type: ignore
        check.state(all(len(s) == 1 for s in glyphs))
        check.state(len(glyphs) == 2)

        self._glyphs = glyphs
        self._l_glyph, self._r_glyph = glyphs

        self._escape_returned_doubles = escape_returned_doubles
        self._compact = compact

        self._l_double_glyph = self._l_glyph * 2
        self._r_double_glyph = self._r_glyph * 2
        self._double_glyphs = (self._l_double_glyph, self._r_double_glyph)

        self._l_escaped_glyph = re.escape(self._l_glyph)
        self._r_escaped_glyph = re.escape(self._r_glyph)

        self._l_glyph_pat = re.compile(r'(%s)' % (self._l_escaped_glyph * 2,))  # noqa
        self._r_glyph_pat = re.compile(r'(%s)' % (self._r_escaped_glyph * 2,))  # noqa

        self._single_glyph_pat = re.compile(r'(%s[^%s]*?%s)' % (self._l_escaped_glyph, self._r_escaped_glyph, self._r_escaped_glyph))  # noqa

    def split(self, s: str) -> ta.List[ta.Union[GlyphSplitMatch, str]]:
        ps = self._l_glyph_pat.split(s)
        ps = [p[::-1] for p in ps for p in reversed(self._r_glyph_pat.split(p[::-1]))]

        ret = []  # type: ignore

        def append_ret(o):
            if self._compact and isinstance(o, str) and ret and isinstance(ret[-1], str):
                ret[-1] = ret[-1] + o
            else:
                ret.append(o)

        for p in ps:
            if p in self._double_glyphs:
                if self._escape_returned_doubles:
                    p = self._glyphs[p == self._r_double_glyph]
                append_ret(p)
                continue

            ms = list(self._single_glyph_pat.finditer(p))
            if not ms:
                append_ret(p)
                continue

            l = 0
            for m in ms:
                if m.start() != l:
                    append_ret(p[l:m.start()])
                append_ret(GlyphSplitMatch(self._l_glyph, p[m.start() + 1:m.end() - 1], self._r_glyph))
                l = m.end()

            if l < len(p):
                append_ret(p[l:])

        return ret


_PAREN_GLYPH_SPLITTER = GlyphSplitter(('(', ')'), escape_returned_doubles=True, compact=True)
_BRACE_GLYPH_SPLITTER = GlyphSplitter(('{', '}'), escape_returned_doubles=True, compact=True)
_BRACKET_GLYPH_SPLITTER = GlyphSplitter(('[', ']'), escape_returned_doubles=True, compact=True)
_ANGLE_BRACKET_GLYPH_SPLITTER = GlyphSplitter(('<', '>'), escape_returned_doubles=True, compact=True)

glyph_split_parens = _PAREN_GLYPH_SPLITTER.split
glyph_split_braces = _BRACE_GLYPH_SPLITTER.split
glyph_split_brackets = _BRACKET_GLYPH_SPLITTER.split
glyph_split_angle_brackets = _ANGLE_BRACKET_GLYPH_SPLITTER.split


def glyph_split_interpolate(
        split_fn: ta.Callable[[str], ta.Sequence[ta.Union[GlyphSplitMatch, str]]],
        dct: ta.Mapping[str, str],
        s: str,
) -> str:
    if not s:
        return s
    sps = split_fn(s)
    if len(sps) == 1 and isinstance(sps[0], str):
        return sps[0]
    return ''.join(dct[p.s] if isinstance(p, GlyphSplitMatch) else p for p in sps)



import dataclasses as dc
import typing as ta

from ... import lang
from .base import Compression
from .codecs import make_compression_codec
from .codecs import make_compression_lazy_loaded_codec


if ta.TYPE_CHECKING:
    import snappy
else:
    snappy = lang.proxy_import('snappy')


##


@dc.dataclass(frozen=True)
class SnappyCompression(Compression):
    def compress(self, d: bytes) -> bytes:
        return snappy.compress(d)

    def decompress(self, d: bytes) -> bytes:
        return snappy.decompress(d)


##


SNAPPY_CODEC = make_compression_codec('snappy', SnappyCompression)

# @omlish-manifest
_SNAPPY_LAZY_CODEC = make_compression_lazy_loaded_codec(__name__, 'SNAPPY_CODEC', SNAPPY_CODEC)



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
TODO:
 - custom merging by key - replace, append, setdefault, equality enforcement, ...
 - better handling of 'None' as signal to replace or not
 - remove inherited
"""
import typing as ta

from ..types import ConfigMap


##


def build_config_inherited_values(
        m: ta.Mapping[str, ConfigMap],
        *,
        inherits_key: str = 'inherits',
) -> ta.Dict[str, ConfigMap]:
    done: ta.Dict[str, ConfigMap] = {}

    def rec(mk: str) -> ta.Mapping[str, ConfigMap]:
        try:
            return done[mk]
        except KeyError:
            pass

        cur = m[mk]
        try:
            inh_seq = cur[inherits_key]
        except KeyError:
            done[mk] = cur
            return cur

        new = dict(cur)
        for inh in inh_seq:
            inh_dct = rec(inh)
            new.update({
                k: v
                for k, v in inh_dct.items()
                if v is not None
                and cur.get(k) is None
            })

        out = {
            k: v
            for k, v in new.items()
            if k != inherits_key
        }

        done[mk] = out
        return out

    for mk in m:
        rec(mk)
    return done



"""
TODO:
 - QualifiedName
 - hybrid dataclass scheme
 - sqlite without sqlalchemy

==

@td.tableclass([
    IdIntegerPrimaryKey(),
    CreatedAtUpdatedAt(),
])
@dc.dataclass(frozen=True)
class User(lang.Final):
    id: int

    created_at: datetime.datetime
    updated_at: datetime.datetime

    name: str

==

@td.tableclass([
    IdIntegerPrimaryKey(),
    CreatedAtUpdatedAt(),
])
@dc.dataclass(frozen=True)
class BaseTable(lang.Abstract):
    id: int

    created_at: datetime.datetime
    updated_at: datetime.datetime

@td.tableclass('user')
@dc.dataclass(frozen=True)
class User(BaseTable, lang.Final):
    name: str

"""
from ... import dataclasses as dc
from ... import lang
from .elements import Elements


##


@dc.dataclass(frozen=True)
class TableDef(lang.Final):
    name: str
    elements: Elements



# ruff: noqa: UP006 UP007
# @omlish-lite
import abc
import typing as ta

from .addresses import SocketAddress


SocketHandlerFactory = ta.Callable[[SocketAddress, ta.BinaryIO, ta.BinaryIO], 'SocketHandler']


##


class SocketHandler(abc.ABC):
    def __init__(
            self,
            client_address: SocketAddress,
            rfile: ta.BinaryIO,
            wfile: ta.BinaryIO,
    ) -> None:
        super().__init__()

        self._client_address = client_address
        self._rfile = rfile
        self._wfile = wfile

    @abc.abstractmethod
    def handle(self) -> None:
        raise NotImplementedError



# ruff: noqa: N802 N803
import typing as ta

from . import runtime as antlr4


LexerT = ta.TypeVar('LexerT', bound=antlr4.Lexer)
ParserT = ta.TypeVar('ParserT', bound=antlr4.Parser)


def parse(
        buf: str,
        lexer_cls: type[LexerT],
        parser_cls: type[ParserT],
) -> ParserT:
    lexer = lexer_cls(antlr4.InputStream(buf))
    stream = antlr4.CommonTokenStream(lexer)
    stream.fill()
    return parser_cls(stream)



"""
TODO:
 - auto drain_asyncio
"""
import contextvars
import functools
import sys
import typing as ta
import warnings

import pytest
from _pytest.outcomes import Skipped  # noqa
from _pytest.outcomes import XFailed  # noqa

from ..... import check
from ..... import lang
from .._registry import register
from .backends import ASYNC_BACKENDS
from .backends import AsyncsBackend
from .consts import ASYNCS_MARK
from .consts import PARAM_NAME
from .fixtures import CANARY
from .fixtures import AsyncsFixture
from .fixtures import AsyncsTestContext
from .fixtures import is_asyncs_fixture
from .utils import is_async_function
from .utils import is_coroutine_function


if ta.TYPE_CHECKING:
    import anyio
else:
    anyio = lang.proxy_import('anyio')


##


@register
class AsyncsPlugin:
    def __init__(self, backends: ta.Collection[type[AsyncsBackend]] | None = None) -> None:
        super().__init__()

        if backends is None:
            backends = ASYNC_BACKENDS

        bd: dict[str, AsyncsBackend] = {}
        for bc in backends:
            be = bc()
            if not be.is_available():
                continue
            bn = be.name
            check.not_in(bn, bd)
            bd[bn] = be
        self._backends = bd

    def pytest_cmdline_main(self, config):
        if (aio_plugin := sys.modules.get('pytest_asyncio.plugin')):
            # warnings.filterwarnings is clobbered by pytest using warnings.catch_warnings
            def aio_plugin_warn(message, *args, **kwargs):
                if (
                        isinstance(message, pytest.PytestDeprecationWarning) and
                        message.args[0].startswith('The configuration option "asyncio_default_fixture_loop_scope" is unset.')  # noqa
                ):
                    return
                warnings.warn(message, *args, **kwargs)

            aio_plugin.warnings = lang.proxy_import('warnings')  # type: ignore
            aio_plugin.warnings.warn = aio_plugin_warn  # type: ignore

    def pytest_configure(self, config):
        config.addinivalue_line('markers', f'{ASYNCS_MARK}: marks for all async backends')

    def pytest_generate_tests(self, metafunc):
        if (m := metafunc.definition.get_closest_marker(ASYNCS_MARK)) is not None:
            if m.args:
                bns = m.args
            else:
                bns = list(self._backends)
        else:
            return

        for bn in bns:
            be = self._backends[bn]
            be.prepare_for_metafunc(metafunc)

        metafunc.fixturenames.append(PARAM_NAME)
        metafunc.parametrize(PARAM_NAME, bns)

    def pytest_fixture_setup(self, fixturedef, request):
        is_asyncs_test = request.node.get_closest_marker(ASYNCS_MARK) is not None

        kwargs = {name: request.getfixturevalue(name) for name in fixturedef.argnames}

        if not is_asyncs_fixture(fixturedef.func, is_asyncs_test, kwargs):
            return None

        if request.scope != 'function':
            raise RuntimeError('Asyncs fixtures must be function-scope')

        if not is_asyncs_test:
            raise RuntimeError('Asyncs fixtures can only be used by Asyncs tests')

        fixture = AsyncsFixture(
            '<fixture {!r}>'.format(fixturedef.argname),  # noqa
            fixturedef.func,
            kwargs,
        )

        fixturedef.cached_result = (fixture, request.param_index, None)

        return fixture

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_call(self, item):
        if (m := item.get_closest_marker(ASYNCS_MARK)) is None:  # noqa
            if is_async_function(item.obj):
                from _pytest.unittest import UnitTestCase  # noqa
                if isinstance(item.parent, UnitTestCase):
                    # unittest handles these itself.
                    pass
                else:
                    raise Exception(f'{item.nodeid}: async def function and no async plugin specified')

            yield
            return

        bn = item.callspec.params[PARAM_NAME]
        be = self._backends[bn]

        item.obj = self.test_runner_factory(be, item)

        yield

    def test_runner_factory(self, backend: AsyncsBackend, item, testfunc=None):
        if not testfunc:
            testfunc = item.obj

        if not is_coroutine_function(testfunc):
            pytest.fail(f'test function `{item!r}` is marked asyncs but is not async')

        @backend.wrap_runner
        async def _bootstrap_fixtures_and_run_test(**kwargs):
            __tracebackhide__ = True

            test_ctx = AsyncsTestContext(backend)
            test = AsyncsFixture(
                '<test {!r}>'.format(testfunc.__name__),  # noqa
                testfunc,
                kwargs,
                is_test=True,
            )

            contextvars_ctx = contextvars.copy_context()
            contextvars_ctx.run(CANARY.set, 'in correct context')

            async with anyio.create_task_group() as nursery:
                for fixture in test.register_and_collect_dependencies():
                    contextvars_ctx.run(
                        functools.partial(
                            nursery.start_soon,
                            fixture.run,
                            test_ctx,
                            contextvars_ctx,
                            name=fixture.name,
                        ),
                    )

            silent_cancellers = test_ctx.fixtures_with_cancel - test_ctx.fixtures_with_errors

            if silent_cancellers:
                for fixture in silent_cancellers:
                    test_ctx.error_list.append(
                        RuntimeError(f"{fixture.name} cancelled the test but didn't raise an error"),
                    )

            if len(test_ctx.error_list) == 1:
                raise test_ctx.error_list[0]
            elif test_ctx.error_list:
                raise BaseExceptionGroup('errors in async test and async fixtures', test_ctx.error_list)

        return _bootstrap_fixtures_and_run_test



import typing as ta


##


class Separators(ta.NamedTuple):
    comma: str
    colon: str


##


PRETTY_INDENT = 2

PRETTY_SEPARATORS = Separators(', ', ': ')

PRETTY_KWARGS: ta.Mapping[str, ta.Any] = dict(
    indent=PRETTY_INDENT,
)


##


COMPACT_SEPARATORS = Separators(',', ':')

COMPACT_KWARGS: ta.Mapping[str, ta.Any] = dict(
    indent=None,
    separators=COMPACT_SEPARATORS,
)



import dataclasses as dc
import typing as ta

from .base import EagerCodec


@dc.dataclass(frozen=True)
class ChainEagerCodec(EagerCodec[ta.Any, ta.Any]):
    codecs: ta.Sequence[EagerCodec]

    def encode(self, v: ta.Any) -> ta.Any:
        for c in self.codecs:
            v = c.encode(v)
        return v

    def decode(self, v: ta.Any) -> ta.Any:
        for c in reversed(self.codecs):
            v = c.decode(v)
        return v


def chain(*codecs: EagerCodec) -> ChainEagerCodec:
    return ChainEagerCodec(codecs)
